/**
 *#########################################################################
 *
 * A component of the Gatherer application, part of the Greenstone digital
 * library suite from the New Zealand Digital Library Project at the
 * University of Waikato, New Zealand.
 *
 * Author: John Thompson, Greenstone Digital Library, University of Waikato
 *
 * Copyright (C) 1999 New Zealand Digital Library Project
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *########################################################################
 */
package org.greenstone.gatherer.cdm;

/**************************************************************************************
 * Written:      16/07/03
 * Revised:      
 **************************************************************************************/
import org.w3c.dom.*;

/** This class represents a level (for mgpp and lucene collecitons) .
 */
public class Level
    implements DOMProxyListEntry {

    /** The Element this level object will source its information from. */
    private Element element;

    /** Constructor used only during DOMProxyListModel initialization. */
    public Level() {
    }

    /** Normal constructor.
     * @param  element the Element this object will find its data from
     */
    public Level(Element element) {
	this.element = element;
    }

    /** Creation of a brand new Level 
     * @param  name the name of this type as a String
     */
    public Level(String level) {
	this.element = CollectionDesignManager.collect_config.document.createElement(CollectionConfiguration.LEVEL_ELEMENT);
	this.element.setAttribute(CollectionConfiguration.NAME_ATTRIBUTE, level);
    }

    /** Compare two objects for ordering. 
     * @param  object the other Object to compare to
     * @return <0 if this level should be before the given object, 0 if they are equal, and >0 if it should be after
     * since we can only have one instance of a particular level, we don't need to look at the name when comparing.
     */
    public int compareTo(Object object) {
	if(object == null) {
	    return -1;
	}
	// hope we never compare a Level with something that is not a Level!
	return getLevel().compareTo(((Level)object).getLevel());
    }

    /** Factory-like method to allow DOMProxyListModel to generate new entries.
     * @param  element the Element the new level should be based on
     * @return a newly created DOMProxyListEntry for the given element
     */
    public DOMProxyListEntry create(Element element) {
	return new Level(element);
    }

    /** Determine if this level is equavilent to another object.
     * @param  object the other Other to match against
     * @return true if the two are equal, false otherwise
     */
    public boolean equals(Object object) {
	return (compareTo(object) == 0);
    }

    /** Retrieve the element this DOMProxyListEntry is based upon. Specified by the interface.
     * @return the Element in question
     */
    public Element getElement() {
	return element;
    }

    /** Retrieve the level.
     * @return the level as a String
     */
    public String getLevel() {
	return element.getAttribute(CollectionConfiguration.NAME_ATTRIBUTE);
    }
    
   /** Determine is this command has been assigned, either because it already existed in the collection configuration, or because it has been explicitly set by the user. Non-assigned entries imply they have been added by the GLI to ensure consistancy (and avoid NPE's!)
     * @return true if this command has been assigned, false otherwise
     * @see org.greenstone.gatherer.cdm.CollectionConfiguration
     */
    public boolean isAssigned() {
	return (element != null && !element.getAttribute(CollectionConfiguration.ASSIGNED_ATTRIBUTE).equals(CollectionConfiguration.FALSE_STR));
    }
    
    /** Set the assigned state.
     * @param  assigned the desired state of assigned as a boolean
     * @see org.greenstone.gatherer.cdm.CollectionConfiguration
     */
    public void setAssigned(boolean assigned) {
	if(element != null) {
	    element.setAttribute(CollectionConfiguration.ASSIGNED_ATTRIBUTE, (assigned ? CollectionConfiguration.TRUE_STR : CollectionConfiguration.FALSE_STR));
	}
    }

    /** Set the element that this DOMProxyListEntry is base on.
     * @param  element the new Element that this entry should source informatin from
     */
    public void setElement(Element element) {
	this.element = element;
    }

    /** Set the name of this level.
     * @param  name the new name for this level, as a String
     * @see org.greenstone.gatherer.cdm.CollectionConfiguration
     */
    public void setLevel(String level) {
	if(element != null) {
	    element.setAttribute(CollectionConfiguration.NAME_ATTRIBUTE, level);
	}
    }

    /** Produce a text representation of this level.
     * @return a String showing this level
     * @see org.greenstone.gatherer.cdm.CollectionConfiguration
     */
    public String toString() {
	return getLevel();
    }
}
